package pl.ivend.pg2oracle.work;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.List;
import java.util.prefs.Preferences;

import pl.ivend.pg2oracle.work.util.ColumnUtil;
import pl.ivend.pg2oracle.work.util.ConstraintUtil;
import pl.ivend.pg2oracle.work.util.ConvParameters;

public abstract class AbstractDatabaseWork {

	protected static final Object MARKS = "\"";
	protected static final String DOT = ".";
	protected static final String COMMA = ",";
	protected static final String SPACE = " ";
	protected static final String NOT = "NOT";
	protected static final String NULL = "NULL";
	protected static final String DEFAULT = "DEFAULT";

	protected static final String PUBLIC = "public";
	protected static final String ID = "id";

	protected abstract List<ColumnDescription> getColumns(String schemat, String table);

	protected abstract DatabaseType getDatabaseType();

	protected abstract Connection getInstance();

	private List<String> replacesTextDefault;

	public AbstractDatabaseWork(Preferences pref) {
		this.replacesTextDefault = prepareReplacesTextDefault(pref);
	}

	protected void close(PreparedStatement ps) {
		close(ps, null);
	}

	protected void close(PreparedStatement ps, ResultSet rs) {
		try {
			if (rs != null) {
				rs.close();
			}
		} catch (Exception e) {
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
			} catch (Exception e) {
			}
		}
	}

	protected boolean isEmpty(String value) {
		return value == null || value.isEmpty();
	}

	protected boolean isNotEmpty(String value) {
		return !isEmpty(value);
	}

	protected TableDescription newTable(String schemaName, String name) {
		TableDescription table = new TableDescription();
		table.setSchemaName(schemaName);
		table.setName(name);
		table.setColumnList(getColumns(table.getSchemaName(), table.getName()));
		return table;
	}

	protected ViewDescription newView(String schemaName, String name, String definition) {
		ViewDescription view = new ViewDescription();
		view.setSchemaName(schemaName);
		view.setName(name);
		view.setDefinition(prepareDefinitionView(definition));
		return view;
	}

	protected String prepareName(String schemaName, String name, String publicUser) {
		StringBuilder sb = new StringBuilder();
		if (PUBLIC.equals(schemaName)) {
			schemaName = publicUser;
		}
		if (schemaName != null) {
			sb.append(prepareName(schemaName));
			sb.append(DOT);
		}
		sb.append(prepareName(name));
		return sb.toString();
	}

	protected String prepareSequenceName(String schemaName, String sequenceName) {
		return prepareName(schemaName, sequenceName, null);
	}

	protected String prepareSequenceName(String schemaName, String sequenceName, String publicUser) {
		return prepareName(schemaName, sequenceName, publicUser);
	}

	protected String prepareTableName(TableDescription tabele) {
		return prepareName(tabele.getSchemaName(), tabele.getName(), null);
	}

	protected String prepareTableName(TableDescription tabele, String publicUser) {
		return prepareName(tabele.getSchemaName(), tabele.getName(), publicUser);
	}

	protected String prepareViewName(ViewDescription view) {
		return prepareName(view.getSchemaName(), view.getName(), null);
	}

	protected String prepareViewName(ViewDescription view, String publicUser) {
		return prepareName(view.getSchemaName(), view.getName(), publicUser);
	}

	protected String prepareColumnDescription(ColumnDescription column) {
		StringBuilder sb = new StringBuilder();
		sb.append(MARKS);
		sb.append(column.getName().toUpperCase());
		sb.append(MARKS);
		sb.append(SPACE);
		ColumnType columnType = column.getType();
		sb.append(ColumnUtil.convTyp(columnType, getDatabaseType()));
		sb.append(prepareDefault(columnType.getDefaultValue()));
		sb.append(SPACE);
		sb.append(prepareIsNull(column.isNull()));
		return sb.toString();
	}

	protected String prepareColumnsDescription(TableDescription tabele) {
		StringBuilder sb = new StringBuilder();
		int count = 0;
		for (ColumnDescription column : tabele.getColumnList()) {
			count++;
			if (count > 1) {
				sb.append(COMMA);
			}
			sb.append(prepareColumnDescription(column));
		}
		return sb.toString();
	}

	protected ConvParameters prepareConvParameters(ResultSet rs) throws Exception {
		ConvParameters convParameters = new ConvParameters();
		convParameters.setColumnType(rs.getString(2));
		convParameters.setDatabaseType(getDatabaseType());
		convParameters.setDataLength(rs.getInt(5));
		String defaultValue = null;
		String columnName = rs.getString(1);
		if (!ID.equals(columnName)) {
			defaultValue = rs.getString(4);
		}
		convParameters.setDefaultValue(defaultValue);
		convParameters.setNumericPrecision(rs.getInt(6));
		convParameters.setNumericScale(rs.getInt(7));
		convParameters.setReplacesTextDefault(replacesTextDefault);
		return convParameters;
	}

	protected ColumnDescription prepareNewColumnDescription(ResultSet rs) throws Exception {
		ColumnDescription column = new ColumnDescription();
		column.setName(rs.getString(1));
		column.setType(ColumnUtil.convString(prepareConvParameters(rs)));
		column.setNull(ColumnUtil.convBool(rs.getString(3), getDatabaseType()));
		column.setDescription(null);
		return column;
	}

	protected ConstraintDescription prepareNewConstraintDescription(ResultSet rs) throws Exception {
		ConstraintDescription constraint = new ConstraintDescription();
		constraint.setSchemaName(rs.getString(1));
		constraint.setTableName(rs.getString(2));
		constraint.setName(rs.getString(3));
		constraint.setType(ConstraintUtil.convString(rs.getString(4), getDatabaseType()));
		return constraint;
	}

	protected String prepareConstraintDescription(ConstraintDescription constraint, String publicUser) {
		StringBuilder sb = new StringBuilder();
		sb.append(prepareAlterTable(constraint, publicUser));
		switch (constraint.getType()) {
		case PKEY:
			sb.append(preparePrimaryKey(constraint));
			return sb.toString();
		case FKEY:
			sb.append(prepareForeignKey(constraint));
			return sb.toString();
		case UNIQUE:
			sb.append(prepareUniqueConstraint(constraint));
			return sb.toString();
		}
		return null;
	}

	protected String prepareConstraintDescription(ConstraintDescription constraint) {
		return prepareConstraintDescription(constraint, null);
	}

	protected String prepareSelectQuery(TableDescription table) {
		return prepareSelectQuery(table, null);
	}

	protected String prepareSelectQuery(TableDescription table, String publicUser) {
		StringBuilder sb = new StringBuilder();
		sb.append("SELECT ");
		int count = 0;
		for (ColumnDescription column : table.getColumnList()) {
			count++;
			if (count > 1) {
				sb.append(COMMA);
			}
			sb.append(prepareName(column.getName()));
		}
		sb.append(" FROM ");
		sb.append(prepareName(table.getSchemaName(), table.getName(), publicUser));
		return sb.toString();
	}

	protected Object prepareDataObject(Object object) {
		return object;
	}

	protected List<ColumnData> prepareColumnDataList(ResultSet rs, List<ColumnDescription> columnList)
			throws Exception {
		List<ColumnData> columnDataList = new ArrayList<ColumnData>();
		for (ColumnDescription column : columnList) {
			ColumnData columnData = new ColumnData();
			columnData.setColumn(column);
			Object object = rs.getObject(column.getName());
			columnData.setObject(prepareDataObject(object));
			columnDataList.add(columnData);
		}
		return columnDataList;
	}

	protected PreparedStatement preparedStatementQuery(String query, List<ColumnData> columnDataList) throws Exception {
		PreparedStatement ps = getInstance().prepareStatement(query);
		int count = 0;
		for (ColumnData columnData : columnDataList) {
			count++;
			ps.setObject(count, columnData.getObject());
		}
		return ps;
	}

	protected String prepareInsertQuery(TableDescription table) {
		return prepareInsertQuery(table, null);
	}

	protected String prepareInsertQuery(TableDescription table, String publicUser) {
		StringBuilder sb = new StringBuilder();
		StringBuilder i = new StringBuilder();
		StringBuilder v = new StringBuilder();
		int count = 0;
		for (ColumnDescription column : table.getColumnList()) {
			count++;
			if (count > 1) {
				i.append(COMMA);
				v.append(COMMA);
			}
			i.append(prepareName(column.getName()));
			v.append("?");
		}
		sb.append("INSERT INTO ");
		sb.append(prepareName(table.getSchemaName(), table.getName(), publicUser));
		sb.append(" (");
		sb.append(i.toString());
		sb.append(")");
		sb.append(" VALUES (");
		sb.append(v.toString());
		sb.append(")");
		return sb.toString();
	}

	protected String prepareConstraintReferences(ConstraintReference reference, String schemaName) {
		StringBuilder sb = new StringBuilder();
		sb.append("GRANT REFERENCES ON ");
		sb.append(prepareName(reference.getSchemaName()));
		sb.append(DOT);
		sb.append(prepareName(reference.getTableName()));
		sb.append(" TO ");
		sb.append(prepareName(schemaName));
		return sb.toString();
	}

	protected String prepareConstraintReferences(ConstraintReference reference, String schemaName, String publicUser) {
		if (PUBLIC.equals(schemaName)) {
			schemaName = publicUser;
		}
		return prepareConstraintReferences(reference, schemaName);
	}

	protected boolean hasNameAdding(String name, String... names) {
		if (names != null && names.length > 0) {
			return findName(name, names);
		}
		return true;
	}

	private List<String> prepareReplacesTextDefault(String replaces) {
		List<String> replaceList = new ArrayList<String>();
		for (String replace : replaces.split(COMMA)) {
			replaceList.add(replace);
		}
		return replaceList;
	}

	private List<String> prepareReplacesTextDefault(Preferences pref) {
		return prepareReplacesTextDefault(pref.get("replaces_text_default", ""));
	}

	private String prepareDefinitionView(String definition) {
		int index = definition.lastIndexOf(";");
		if (index != -1) {
			return definition.substring(0, index);
		} else {
			return definition;
		}
	}

	private String prepareIsNull(boolean isNull) {
		StringBuilder sb = new StringBuilder();
		if (isNull) {
			sb.append(NULL);
		} else {
			sb.append(NOT);
			sb.append(SPACE);
			sb.append(NULL);
		}
		return sb.toString();
	}

	private String prepareDefault(String defaultValue) {
		StringBuilder sb = new StringBuilder();
		if (defaultValue != null && !defaultValue.isEmpty()) {
			sb.append(SPACE);
			sb.append(DEFAULT);
			sb.append(SPACE);
			sb.append(defaultValue);
		}
		return sb.toString();
	}

	private boolean findName(String name, String... names) {
		for (String n : names) {
			if (name.equals(n)) {
				return true;
			}
		}
		return false;
	}

	private String prepareName(String name) {
		StringBuilder sb = new StringBuilder();
		switch (getDatabaseType()) {
		case ORACLE:
			sb.append(MARKS);
			sb.append(name.toUpperCase());
			sb.append(MARKS);
			break;
		case POSTGRES:
			sb.append(name);
			break;
		}
		return sb.toString();
	}

	private String prepareAlterTable(ConstraintDescription constraint, String publicUser) {
		StringBuilder sb = new StringBuilder();
		sb.append("ALTER TABLE ");
		String schemaName = constraint.getSchemaName();
		if (PUBLIC.equals(schemaName)) {
			schemaName = publicUser;
		}
		if (schemaName != null) {
			sb.append(prepareName(schemaName));
			sb.append(DOT);
		}
		sb.append(prepareName(constraint.getTableName()));
		sb.append(" ADD CONSTRAINT ");
		sb.append(constraint.getName());
		sb.append(" ");
		return sb.toString();
	}

	private String prepareConstraintKeys(List<String> columnList) {
		StringBuilder sb = new StringBuilder();
		int count = 0;
		for (String column : columnList) {
			count++;
			if (count > 1) {
				sb.append(COMMA);
			}
			sb.append(column);
		}
		return sb.toString();
	}

	private String prepareConstraintReferences(List<ConstraintReference> referenceList) {
		StringBuilder sb = new StringBuilder();
		int count = 0;
		for (ConstraintReference reference : referenceList) {
			count++;
			if (count > 1) {
				sb.append(COMMA);
			}
			sb.append(prepareName(reference.getSchemaName()));
			sb.append(DOT);
			sb.append(prepareName(reference.getTableName()));
			sb.append("(");
			sb.append(reference.getColumnName());
			sb.append(")");
		}
		return sb.toString();
	}

	private String preparePrimaryKey(ConstraintDescription constraint) {
		StringBuilder sb = new StringBuilder();
		sb.append("PRIMARY KEY (");
		sb.append(prepareConstraintKeys(constraint.getKeyList()));
		sb.append(")");
		return sb.toString();
	}

	private String prepareForeignKey(ConstraintDescription constraint) {
		StringBuilder sb = new StringBuilder();
		sb.append("FOREIGN KEY (");
		sb.append(prepareConstraintKeys(constraint.getKeyList()));
		sb.append(") REFERENCES ");
		sb.append(prepareConstraintReferences(constraint.getReferenceList()));
		return sb.toString();
	}

	private String prepareUniqueConstraint(ConstraintDescription constraint) {
		StringBuilder sb = new StringBuilder();
		sb.append("UNIQUE (");
		sb.append(prepareConstraintKeys(constraint.getKeyList()));
		sb.append(")");
		return sb.toString();
	}

}
