package pl.ivend.pg2oracle.work;

import java.util.ArrayList;
import java.util.List;
import java.util.prefs.Preferences;

import pl.ivend.pg2oracle.AppLogger;
import pl.ivend.pg2oracle.work.oracle.OracleWork;
import pl.ivend.pg2oracle.work.postgresql.PostgresWork;

public final class Convert {

	private static final String POSTGRESQL = "postgres";
	private static final String ORACLE = "oracle";

	private static final String COPY_SCHEMA = "Kopiowanie schematów... ";
	private static final String COPY_TABLE = "Kopiowanie tabel... ";
	private static final String COPY_SEQUENCE = "Kopiowanie sekwencji...";
	private static final String UPDATE_SEQUENCE = "Aktualizacja sekwencji...";
	private static final String COPY_VIEW = "Kopiowanie widoków...";
	private static final String COPY_DATA = "Kopiowanie danych...";
	private static final String COPY_PRIMARY_KEY = "Kopiowanie kluczy głównych...";;
	private static final String COPY_CONSTRAINT = "Kopiowanie powiązań...";
	private static final String COPY_TRIGGER = "Kopiowanie trigerów...";
	private static final String OK = "OK";

	private List<String> schemaList;
	private List<TableDescription> tableList = new ArrayList<TableDescription>();
	private IDatabase postgresql;
	private IDatabase oracle;

	private ConvertArgs args;

	public Convert(ConvertArgs args, Preferences pref) {
		this.args = args;
		postgresql = new PostgresWork(pref.node(POSTGRESQL));
		oracle = new OracleWork(pref.node(ORACLE));
	}

	private void addTableToList(TableDescription table) {
		tableList.add(table);
	}

	private void copyingSchemas(IDatabase source, IDatabase dest) {
		schemaList = source.getSchemas(args.getSchemaNames());
		for (String schemaName : schemaList) {
			boolean isSchema = dest.isSchema(schemaName);
			if (!isSchema) {
				dest.createSchema(schemaName);
			}
		}
	}

	private void copyingSequences(IDatabase source, IDatabase dest) {
		for (String schemaName : schemaList) {
			List<String> sequenceList = source.getSequences(schemaName);
			for (String sequenceName : sequenceList) {
				boolean isSequence = dest.isSequence(schemaName, sequenceName);
				if (!isSequence) {
					SequenceDescription sequence = source.getSequence(schemaName, sequenceName);
					dest.createSequence(sequence);
				}
			}
		}
	}

	private void updateSequences(IDatabase source, IDatabase dest) {
		for (String schemaName : schemaList) {
			List<String> sequenceList = source.getSequences(schemaName);
			for (String sequenceName : sequenceList) {
				boolean isSequence = dest.isSequence(schemaName, sequenceName);
				if (isSequence) {
					SequenceDescription sequence = source.getSequence(schemaName, sequenceName);
					dest.updateSequence(sequence);
				}
			}
		}
	}

	private void copyingTables(IDatabase source, IDatabase dest) {
		for (String schemaName : schemaList) {
			List<TableDescription> tableList = source.getTables(schemaName, args.getTableNames());
			for (TableDescription table : tableList) {
				addTableToList(table);
				boolean isTable = dest.isTable(table.getSchemaName(), table.getName());
				if (!isTable) {
					dest.createTable(table);
				}
			}
		}
	}

	private void copyingViews(IDatabase source, IDatabase dest) {
		for (String schemaName : schemaList) {
			List<ViewDescription> viewList = source.getViews(schemaName);
			for (ViewDescription view : viewList) {
				boolean isView = dest.isView(view.getSchemaName(), view.getName());
				if (!isView) {
					dest.createView(view);
				}
			}
		}
	}

	private void copyingTriggers(IDatabase source, IDatabase dest) {
		for (String schemaName : schemaList) {
			List<TriggerDescription> triggerList = source.getTriggers(schemaName);
			for (TriggerDescription trigger : triggerList) {
				boolean isTrigger = dest.isTrigger(trigger.getSchemaName(), trigger.getName());
				if (!isTrigger) {
					dest.createTrigger(trigger);
				}
			}
		}
	}

	private void copyingDatas(IDatabase source, final IDatabase dest, TableDescription table) {
		source.getDataAsync(table, new IDataAsync() {

			@Override
			public void insertData(TableDescription table, List<ColumnData> dataList) {
				dest.insertData(table, dataList);
			}
		});
	}

	private void copyingDatas(IDatabase source, IDatabase dest) {
		for (TableDescription table : tableList) {
			String schemaName = table.getSchemaName();
			String tableName = table.getName();
			boolean isTable = dest.isTable(schemaName, tableName);
			if (isTable) {
				copyingDatas(source, dest, table);
			}
		}
	}

	private boolean isForeignKey(ConstraintDescription constraint) {
		return ConstraintType.FKEY.equals(constraint.getType());
	}

	private void addConstraintReferencesGrants(IDatabase dest, ConstraintDescription constraint) {
		if (isForeignKey(constraint)) {
			dest.addConstraintReferencesGrants(constraint);
		}
	}

	private void createConstraint(IDatabase dest, ConstraintDescription constraint) {
		dest.createConstraint(constraint);
	}

	private void copyingConstraints(IDatabase source, IDatabase dest, String schemaName, String tableName,
			List<ConstraintDescription> constraintList) {
		for (ConstraintDescription constraint : constraintList) {
			boolean isConstraint = dest.isConstraint(schemaName, tableName, constraint.getName());
			if (!isConstraint) {
				addConstraintReferencesGrants(dest, constraint);
				createConstraint(dest, constraint);
			}
		}
	}

	private void copyingConstraints(IDatabase source, IDatabase dest, boolean isCopyingPrimaryKey) {
		for (TableDescription table : tableList) {
			String schemaName = table.getSchemaName();
			String tableName = table.getName();
			boolean isTable = dest.isTable(schemaName, tableName);
			if (isTable) {
				List<ConstraintDescription> constraintList = null;
				if (isCopyingPrimaryKey) {
					constraintList = source.getPrimaryKeys(schemaName, tableName);
				} else {
					constraintList = source.getConstraints(schemaName, tableName);
				}
				if (constraintList != null) {
					copyingConstraints(source, dest, schemaName, tableName, constraintList);
				}
			}
		}
	}

	public void postgresToOracle() {
		sourceToDestination(postgresql, oracle);
	}

	public void oracleToPostgres() {
		sourceToDestination(oracle, postgresql);
	}

	public void sourceToDestination(IDatabase source, IDatabase dest) {
		if (args.isSchemas()) {
			AppLogger.get().info(COPY_SCHEMA);
			copyingSchemas(source, dest);
			AppLogger.get().info(OK);
		}
		if (args.isSequences()) {
			AppLogger.get().info(COPY_SEQUENCE);
			copyingSequences(source, dest);
			AppLogger.get().info(OK);
		}
		if (args.isTables()) {
			AppLogger.get().info(COPY_TABLE);
			copyingTables(source, dest);
			AppLogger.get().info(OK);
		}
		if (args.isDatas()) {
			AppLogger.get().info(COPY_DATA);
			copyingDatas(source, dest);
			AppLogger.get().info(OK);
		}
		if (args.isPrimaryKeys()) {
			AppLogger.get().info(COPY_PRIMARY_KEY);
			copyingConstraints(source, dest, true);
			AppLogger.get().info(OK);
		}
		if (args.isConstraints()) {
			AppLogger.get().info(COPY_CONSTRAINT);
			copyingConstraints(source, dest, false);
			AppLogger.get().info(OK);
		}
		if (args.isViews()) {
			AppLogger.get().info(COPY_VIEW);
			copyingViews(source, dest);
			AppLogger.get().info(OK);
		}
		if (args.isTriggers()) {
			AppLogger.get().info(COPY_TRIGGER);
			copyingTriggers(source, dest);
			AppLogger.get().info(OK);
		}
		if (args.isSequencesUpdate()) {
			AppLogger.get().info(UPDATE_SEQUENCE);
			updateSequences(source, dest);
			AppLogger.get().info(OK);
		}
	}

}
