package pl.ivend.pg2oracle.work.oracle;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.List;
import java.util.prefs.Preferences;

import pl.ivend.pg2oracle.AppLogger;
import pl.ivend.pg2oracle.work.AbstractDatabaseWork;
import pl.ivend.pg2oracle.work.ColumnData;
import pl.ivend.pg2oracle.work.ColumnDescription;
import pl.ivend.pg2oracle.work.ConstraintDescription;
import pl.ivend.pg2oracle.work.ConstraintReference;
import pl.ivend.pg2oracle.work.DatabaseType;
import pl.ivend.pg2oracle.work.IDataAsync;
import pl.ivend.pg2oracle.work.IDatabase;
import pl.ivend.pg2oracle.work.SequenceDescription;
import pl.ivend.pg2oracle.work.TableDescription;
import pl.ivend.pg2oracle.work.TriggerDescription;
import pl.ivend.pg2oracle.work.ViewDescription;
import pl.ivend.pg2oracle.work.util.ConstraintUtil;

public final class OracleWork extends AbstractDatabaseWork implements IDatabase {

	// private static final int MAX_LENGTH_NAME_ORACLE = 30;

	private Connection conn;
	private String publicUser;

	public OracleWork(Preferences pref) {
		super(pref);
		try {
			Class.forName("oracle.jdbc.driver.OracleDriver");
			String hostname = pref.get("hostname", "127.0.0.1");
			String port = pref.get("port", "1521");
			String database = pref.get("database", "XE");
			String username = pref.get("username", "sys");
			String password = pref.get("password", "");
			publicUser = pref.get("publicuser", "sys");
			conn = DriverManager.getConnection("jdbc:oracle:thin:@" + hostname + ":" + port + ":" + database, username,
					password);
		} catch (Exception e) {
			AppLogger.get().error(e);
		}
	}

	@Override
	public Connection getInstance() {
		return conn;
	}

	@Override
	public List<String> getSchemas(String... schemas) {
		List<String> schemaList = new ArrayList<String>();
		return schemaList;
	}

	@Override
	public boolean isSchema(String schemaName) {
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select username from all_users where username = ?";
		try {
			if (PUBLIC.equals(schemaName)) {
				return true;
			}
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			rs = ps.executeQuery();
			if (rs.next()) {
				return true;
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return false;
	}

	@Override
	public void createSchema(String schemaName) {
		PreparedStatement ps = null;
		String query = null;
		try {
			if (PUBLIC.equals(schemaName)) {
				return;
			}
			schemaName = schemaName.toUpperCase();
			query = preapreCreateUser(schemaName);
			AppLogger.get().info(query);
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			query = preapreLockUser(schemaName);
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			AppLogger.get().info(query);
			query = preapreGrantUnlimitedTableToUser(schemaName);
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			AppLogger.get().info(query);
			conn.commit();
		} catch (Exception e1) {
			AppLogger.get().error(query, e1);
			try {
				conn.rollback();
			} catch (Exception e2) {
				AppLogger.get().error(e2);
			}
		} finally {
			close(ps);
		}
	}

	@Override
	public SequenceDescription getSequence(String schemaName, String sequenceName) {
		PreparedStatement ps = null;
		ResultSet rs = null;
		SequenceDescription sequence = new SequenceDescription();
		sequence.setSchemaName(schemaName);
		sequence.setName(sequenceName);
		final String query = "select last_number,increment_by,max_value,min_value from all_sequences where sequence_owner = ? and sequence_name = ?";
		try {
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName);
			ps.setString(2, sequenceName);
			rs = ps.executeQuery();
			if (rs.next()) {
				sequence.setValueLast(rs.getLong(1));
				sequence.setValueStart(sequence.getValueLast());
				sequence.setIncrementBy(rs.getLong(2));
				sequence.setValueMax(rs.getLong(3));
				sequence.setValueMin(rs.getLong(4));
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return sequence;
	}

	@Override
	public List<String> getSequences(String schemaName) {
		List<String> sequenceList = new ArrayList<String>();
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select sequence_name from all_sequences where sequence_owner = ?";
		try {
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			rs = ps.executeQuery();
			while (rs.next()) {
				String sequenceName = rs.getString(1);
				sequenceList.add(sequenceName);
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return sequenceList;
	}

	@Override
	public boolean isSequence(String schemaName, String sequenceName) {
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select sequence_name from all_sequences where sequence_owner = ? and sequence_name = ?";
		try {
			if (PUBLIC.equals(schemaName)) {
				schemaName = publicUser;
			}
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			ps.setString(2, sequenceName.toUpperCase());
			rs = ps.executeQuery();
			if (rs.next()) {
				return true;
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return false;
	}

	@Override
	public void createSequence(SequenceDescription sequence) {
		PreparedStatement ps = null;
		String query = null;
		try {
			StringBuilder sb = new StringBuilder();
			sb.append("CREATE ");
			sb.append(prepareSequence(sequence));
			query = sb.toString();
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			AppLogger.get().info(query);
			conn.commit();
		} catch (Exception e) {
			AppLogger.get().error(query, e);
			try {
				conn.rollback();
			} catch (Exception e2) {
				AppLogger.get().error(e2);
			}
		} finally {
			close(ps);
		}
	}

	@Override
	public void updateSequence(SequenceDescription sequence) {
		PreparedStatement ps = null;
		String query = null;
		try {
			StringBuilder sb = new StringBuilder();
			sb.append("ALTER ");
			sb.append(prepareSequence(sequence));
			query = sb.toString();
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			AppLogger.get().info(query);
			conn.commit();
		} catch (Exception e) {
			AppLogger.get().error(query, e);
			try {
				conn.rollback();
			} catch (Exception e2) {
				AppLogger.get().error(e2);
			}
		} finally {
			close(ps);
		}
	}

	@Override
	public List<TableDescription> getTables(String schemaName, String... tables) {
		List<TableDescription> tabeleList = new ArrayList<TableDescription>();
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select table_name from all_tables where owner=?";
		try {
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			rs = ps.executeQuery();
			while (rs.next()) {
				String tableName = rs.getString(1);
				if (hasNameAdding(tableName, tables)) {
					TableDescription table = newTable(schemaName, tableName);
					tabeleList.add(table);
				}
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return tabeleList;
	}

	@Override
	public boolean isTable(String schemaName, String tableName) {
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select table_name from all_tables where owner=? and table_name=?";
		try {
			if (PUBLIC.equals(schemaName)) {
				schemaName = publicUser;
			}
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			ps.setString(2, tableName.toUpperCase());
			rs = ps.executeQuery();
			if (rs.next()) {
				return true;
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return false;
	}

	@Override
	public void createTable(TableDescription tabele) {
		PreparedStatement ps = null;
		String query = null;
		try {
			String tableName = prepareTableName(tabele, publicUser);
			String columnsDescription = prepareColumnsDescription(tabele);
			StringBuilder sb = new StringBuilder();
			sb.append("create table ");
			sb.append(tableName);
			sb.append(" (");
			sb.append(columnsDescription);
			sb.append(")");
			query = sb.toString();
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			AppLogger.get().info(query);
			conn.commit();
		} catch (Exception e) {
			AppLogger.get().error(query, e);
			try {
				conn.rollback();
			} catch (Exception e2) {
				AppLogger.get().error(e2);
			}
		} finally {
			close(ps);
		}
	}

	@Override
	public List<ViewDescription> getViews(String schemaName) {
		List<ViewDescription> viewList = new ArrayList<ViewDescription>();
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select view_name,text from all_views where owner=?";
		try {
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			rs = ps.executeQuery();
			while (rs.next()) {
				ViewDescription view = newView(schemaName, rs.getString(1), rs.getString(2));
				viewList.add(view);
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return viewList;
	}

	@Override
	public boolean isView(String schemaName, String viewName) {
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select view_name from all_views where owner=? and view_name = ?";
		try {
			if (PUBLIC.equals(schemaName)) {
				schemaName = publicUser;
			}
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			ps.setString(2, viewName.toUpperCase());
			rs = ps.executeQuery();
			if (rs.next()) {
				return true;
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return false;
	}

	@Override
	public void createView(ViewDescription view) {
		PreparedStatement ps = null;
		String query = null;
		try {
			String viewName = prepareViewName(view, publicUser);
			StringBuilder sb = new StringBuilder();
			sb.append("create or replace view ");
			sb.append(viewName);
			sb.append(" as ");
			sb.append(view.getDefinition());
			query = sb.toString();
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			AppLogger.get().info(query);
			conn.commit();
		} catch (Exception e) {
			AppLogger.get().error(query, e);
			try {
				conn.rollback();
			} catch (Exception e2) {
				AppLogger.get().error(e2);
			}
		} finally {
			close(ps);
		}
	}

	@Override
	public void getDataAsync(TableDescription table, IDataAsync dataAsync) {
		PreparedStatement ps = null;
		ResultSet rs = null;
		String query = null;
		try {
			query = prepareSelectQuery(table, publicUser);
			ps = conn.prepareStatement(query);
			rs = ps.executeQuery();
			while (rs.next()) {
				List<ColumnDescription> columnList = table.getColumnList();
				dataAsync.insertData(table, prepareColumnDataList(rs, columnList));
			}

		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
	}

	@Override
	public void insertData(TableDescription table, List<ColumnData> dataList) {
		PreparedStatement ps = null;
		String query = null;
		try {
			query = prepareInsertQuery(table, publicUser);
			ps = preparedStatementQuery(query, dataList);
			ps.executeUpdate();
			AppLogger.get().info(query);
			conn.commit();
		} catch (Exception e) {
			AppLogger.get().error(query, e);
			try {
				conn.rollback();
			} catch (Exception e2) {
				AppLogger.get().error(e2);
			}
		} finally {
			close(ps);
		}
	}

	@Override
	public List<ConstraintDescription> getPrimaryKeys(String schemaName, String tableName) {
		final String query = "select owner,table_name,constraint_name,constraint_type from all_constraints where constraint_type = ? and owner = ? and table_name = ?";
		List<ConstraintDescription> constraintList = getConstraints(schemaName, tableName, query);
		return constraintList;
	}

	@Override
	public List<ConstraintDescription> getConstraints(String schemaName, String tableName) {
		final String query = "select owner,table_name,constraint_name,constraint_type from all_constraints where constraint_type <> ? and owner = ? and table_name = ?";
		List<ConstraintDescription> constraintList = getConstraints(schemaName, tableName, query);
		return constraintList;
	}

	@Override
	public boolean isConstraint(String schemaName, String tableName, String constraintName) {
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select constraint_name,constraint_type from all_constraints where owner = ? and table_name = ? and constraint_name = ?";
		try {
			if (PUBLIC.equals(schemaName)) {
				schemaName = publicUser;
			}
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			ps.setString(2, tableName.toUpperCase());
			// ps.setString(3, prepareConstraintName(constraintName));
			ps.setString(3, constraintName.toUpperCase());
			rs = ps.executeQuery();
			if (rs.next()) {
				return true;
			}
		} catch (Exception e) {
			AppLogger.get().error(e);
		} finally {
			close(ps, rs);
		}
		return false;
	}

	@Override
	public void createConstraint(ConstraintDescription constraint) {
		PreparedStatement ps = null;
		String query = null;
		try {
			query = prepareConstraintDescription(constraint, publicUser);
			if (isEmpty(query)) {
				return;
			}
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			AppLogger.get().info(query);
			conn.commit();
		} catch (Exception e) {
			AppLogger.get().error(query, e);
			try {
				conn.rollback();
			} catch (Exception e2) {
				AppLogger.get().error(e2);
			}
		} finally {
			close(ps);
		}
	}

	@Override
	public void addConstraintReferencesGrants(ConstraintDescription constraint) {
		List<ConstraintReference> referenceList = constraint.getReferenceList();
		if (referenceList != null) {
			for (ConstraintReference reference : referenceList) {
				addConstraintReferencesGrant(reference, constraint.getSchemaName());
			}
		}
	}

	@Override
	public List<TriggerDescription> getTriggers(String schemaName) {
		// TODO Auto-generated method stub
		return new ArrayList<TriggerDescription>();
	}

	@Override
	public boolean isTrigger(String schemaName, String name) {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public void createTrigger(TriggerDescription trigger) {
		// TODO Auto-generated method stub

	}

	@Override
	protected DatabaseType getDatabaseType() {
		return DatabaseType.ORACLE;
	}

	@Override
	protected List<ColumnDescription> getColumns(String schemaName, String tabeleName) {
		List<ColumnDescription> columnList = new ArrayList<ColumnDescription>();
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select column_name, data_type, nullable, data_default, data_length, data_precision, data_scale from all_tab_columns where owner=? and table_name=?";
		try {
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			ps.setString(2, tabeleName.toUpperCase());
			rs = ps.executeQuery();
			while (rs.next()) {
				columnList.add(prepareNewColumnDescription(rs));
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return columnList;
	}

	private String prepareSequence(SequenceDescription sequence) {
		StringBuilder sb = new StringBuilder();
		sb.append("SEQUENCE ");
		sb.append(prepareSequenceName(sequence.getSchemaName(), sequence.getName(), publicUser));
		sb.append(" MINVALUE ");
		sb.append(sequence.getValueMin());
		sb.append(" START WITH ");
		sb.append(sequence.getValueStart());
		sb.append(" INCREMENT BY ");
		sb.append(sequence.getIncrementBy());
		sb.append(" CACHE 20");
		return sb.toString();
	}

	private void addConstraintReferencesGrant(ConstraintReference reference, String schemaName) {
		PreparedStatement ps = null;
		String query = null;
		try {
			query = prepareConstraintReferences(reference, schemaName, publicUser);
			if (isEmpty(query)) {
				return;
			}
			ps = conn.prepareStatement(query);
			ps.executeUpdate();
			AppLogger.get().info(query);
			conn.commit();
		} catch (Exception e) {
			AppLogger.get().error(query, e);
			try {
				conn.rollback();
			} catch (Exception e2) {
				AppLogger.get().error(e2);
			}
		} finally {
			close(ps);
		}
	}

	private List<ConstraintDescription> getConstraints(String schemaName, String tableName, String query) {
		List<ConstraintDescription> constraintList = new ArrayList<ConstraintDescription>();
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			ps = conn.prepareStatement(query);
			ps.setString(1, ConstraintUtil.getPrimaryKey(getDatabaseType()));
			ps.setString(2, schemaName.toUpperCase());
			ps.setString(3, tableName.toUpperCase());
			rs = ps.executeQuery();
			while (rs.next()) {
				ConstraintDescription newConstraint = prepareNewConstraintDescription(rs);
				String constraintName = newConstraint.getName();
				newConstraint.setKeyList(getConstraintKeys(schemaName, tableName, constraintName));
				newConstraint.setReferenceList(getConstraintReferences(schemaName, constraintName));
				constraintList.add(newConstraint);
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return constraintList;
	}

	private String preapreCreateUser(String schemaName) {
		StringBuilder sb = new StringBuilder();
		sb.append("CREATE USER ");
		sb.append(schemaName);
		sb.append(" identified by ");
		sb.append(schemaName);
		return sb.toString();
	}

	private String preapreLockUser(String schemaName) {
		StringBuilder sb = new StringBuilder();
		sb.append("ALTER USER ");
		sb.append(schemaName);
		sb.append(" ACCOUNT LOCK");
		return sb.toString();
	}

	private String preapreGrantUnlimitedTableToUser(String schemaName) {
		StringBuilder sb = new StringBuilder();
		sb.append("GRANT UNLIMITED TABLESPACE TO ");
		sb.append(schemaName);
		return sb.toString();
	}

	// private String prepareConstraintName(String constraintName) {
	// int length = constraintName.length();
	// boolean isMaxLength = (length > MAX_LENGTH_NAME_ORACLE);
	// if (isMaxLength) {
	// int startIndex = length - MAX_LENGTH_NAME_ORACLE;
	// constraintName = constraintName.substring(startIndex, length);
	// }
	// return constraintName.toUpperCase();
	// }

	private List<String> getConstraintKeys(String schemaName, String tableName, String constraintName) {
		List<String> constraintColumnList = new ArrayList<String>();
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select column_name from all_cons_columns where owner = ? and table_name = ? and constraint_name = ?";
		try {
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName.toUpperCase());
			ps.setString(2, tableName.toUpperCase());
			ps.setString(3, constraintName.toUpperCase());
			rs = ps.executeQuery();
			while (rs.next()) {
				constraintColumnList.add(rs.getString(1));
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return constraintColumnList;
	}

	private List<ConstraintReference> getConstraintReferences(String schemaName, String constraintName) {
		List<ConstraintReference> constraintReferenceList = new ArrayList<ConstraintReference>();
		PreparedStatement ps = null;
		ResultSet rs = null;
		final String query = "select c.owner, c.table_name, c.column_name from all_constraints a, all_constraints b, all_cons_columns c where a.constraint_name = b.r_constraint_name and a.owner = b.r_owner and b.owner = ? and b.constraint_name = ? and a.constraint_name = c.constraint_name";
		try {
			ps = conn.prepareStatement(query);
			ps.setString(1, schemaName);
			ps.setString(2, constraintName);
			rs = ps.executeQuery();
			while (rs.next()) {
				ConstraintReference reference = new ConstraintReference();
				reference.setSchemaName(rs.getString(1));
				reference.setTableName(rs.getString(2));
				reference.setColumnName(rs.getString(3));
				constraintReferenceList.add(reference);
			}
		} catch (Exception e) {
			AppLogger.get().error(query, e);
		} finally {
			close(ps, rs);
		}
		return constraintReferenceList;
	}

}
