package pl.ivend.pg2oracle.work.util;

import java.util.List;

import pl.ivend.pg2oracle.work.ColumnType;
import pl.ivend.pg2oracle.work.ColumnType.Typ;
import pl.ivend.pg2oracle.work.DatabaseType;
import pl.ivend.pg2oracle.work.exception.ColumnTypeException;
import pl.ivend.pg2oracle.work.exception.DatabaseTypeException;

public final class ColumnUtil {

	private static final String COMMA = ",";
	private static final String APOSTROPHE = "'";
	private static final String COLON = ":";
	private static final Integer MAX_LENGHT_VARCHAR = 4000;

	private ColumnUtil() {
		// TODO Auto-generated constructor stub
	}

	public static boolean convBool(String bool, DatabaseType database) {
		switch (database) {
		case POSTGRES:
			if ("YES".equals(bool)) {
				return true;
			}
			break;
		case ORACLE:
			if ("Y".equals(bool)) {
				return true;
			}
			break;
		}
		return false;
	}

	private static String prepareBooleanTypeDefault(String defaultValue) {
		if ("true".equals(defaultValue)) {
			return "1";
		} else if ("false".equals(defaultValue)) {
			return "0";
		} else {
			return null;
		}
	}

	private static String prepareCharacterTypeDefault(String defaultValue, int startIndex, String separator) {
		int stopIndex = defaultValue.lastIndexOf(separator);
		if (stopIndex > startIndex) {
			return defaultValue.substring(startIndex, stopIndex + 1);
		} else {
			return defaultValue;
		}
	}

	private static String prepareCharacterTypeDefault(String defaultValue, String separator) {
		int startIndex = defaultValue.indexOf(separator);
		if (startIndex != -1) {
			return prepareCharacterTypeDefault(defaultValue, startIndex, separator);
		} else {
			return defaultValue;
		}
	}

	private static String replaceStringDefault(String value, List<String> replaces) {
		for (String replace : replaces) {
			String[] token = replace.split(COLON);
			value = value.replace(token[0], token[1]);
		}
		return value;
	}

	private static String prepareCharacterTypeDefault(ConvParameters convParameters) {
		String defaultValue = convParameters.getDefaultValue();
		if (defaultValue != null) {
			String prepareDefault = prepareCharacterTypeDefault(defaultValue, APOSTROPHE);
			return replaceStringDefault(prepareDefault, convParameters.getReplacesTextDefault());
		} else {
			return null;
		}
	}

	private static String prepareDateTypeDefault(String defaultValue) {
		if ("now()".equals(defaultValue)) {
			return "sysdate";
		} else {
			return null;
		}
	}

	private static ColumnType newColumnType(Typ type, Integer dataLength, Integer numericPrecision,
			Integer numericScale, String defaultValue) {
		ColumnType columnType = new ColumnType();
		columnType.setType(type);
		columnType.setDataLength(dataLength);
		columnType.setNumericPrecision(numericPrecision);
		columnType.setNumericScale(numericScale);
		columnType.setDefaultValue(defaultValue);
		return columnType;
	}

	private static ColumnType newColumnType(Typ type, Integer dataLength) {
		return newColumnType(type, dataLength, null, null, null);
	}

	private static ColumnType newColumnType(Typ type, String defaultValue) {
		return newColumnType(type, null, null, null, defaultValue);
	}

	private static ColumnType newColumnType(Typ type) {
		return newColumnType(type, null, null, null, null);
	}

	private static ColumnType newColumnType(Typ type, Integer dataLength, String defaultValue) {
		return newColumnType(type, dataLength, null, null, defaultValue);
	}

	private static ColumnType newColumnType(Typ type, Integer numericPrecision, Integer numericScale,
			String defaultValue) {
		return newColumnType(type, null, numericPrecision, numericScale, defaultValue);
	}

	private static ColumnType convPostgresString(ConvParameters convParameters) {
		String columnType = convParameters.getColumnType();
		String defaultValue = convParameters.getDefaultValue();
		Integer dataLength = convParameters.getDataLength();
		Integer numericPrecision = convParameters.getNumericPrecision();
		Integer numericScale = convParameters.getNumericScale();
		if ("integer".equals(columnType)) {
			return newColumnType(Typ.NUMBER, defaultValue);
		} else if ("smallint".equals(columnType)) {
			return newColumnType(Typ.NUMBER, defaultValue);
		} else if ("text".equals(columnType)) {
			return newColumnType(Typ.CLOB, prepareCharacterTypeDefault(convParameters));
		} else if ("character varying".equals(columnType)) {
			return newColumnType(Typ.VARCHAR, dataLength, prepareCharacterTypeDefault(convParameters));
		} else if ("name".equals(columnType)) {
			return newColumnType(Typ.VARCHAR, 255);
		} else if ("USER-DEFINED".equals(columnType)) {
			return newColumnType(Typ.VARCHAR, 255);
		} else if ("oid".equals(columnType)) {
			return newColumnType(Typ.NUMBER, defaultValue);
		} else if ("character".equals(columnType)) {
			return newColumnType(Typ.CHAR, dataLength, prepareCharacterTypeDefault(convParameters));
		} else if ("bytea".equals(columnType)) {
			return newColumnType(Typ.BLOB, defaultValue);
		} else if ("boolean".equals(columnType)) {
			return newColumnType(Typ.NUMBER, 1, 0, prepareBooleanTypeDefault(defaultValue));
		} else if ("date".equals(columnType)) {
			return newColumnType(Typ.DATE, prepareDateTypeDefault(defaultValue));
		} else if (columnType.startsWith("timestamp")) {
			return newColumnType(Typ.TIMESTAMP, prepareDateTypeDefault(defaultValue));
		} else if (columnType.startsWith("time")) {
			return newColumnType(Typ.DATE, prepareDateTypeDefault(defaultValue));
		} else if ("numeric".equals(columnType)) {
			return newColumnType(Typ.NUMBER, numericPrecision, numericScale, defaultValue);
		} else if (columnType.startsWith("double")) {
			return newColumnType(Typ.NUMBER, defaultValue);
		} else if ("bigint".equals(columnType)) {
			return newColumnType(Typ.NUMBER, defaultValue);
		} else if ("real".equals(columnType)) {
			return newColumnType(Typ.NUMBER, defaultValue);
		} else {
			return newColumnType(Typ.UNKNOWN);
		}
	}

	private static ColumnType convOracleString(ConvParameters convParameters) {
		String columnType = convParameters.getColumnType();
		String defaultValue = convParameters.getDefaultValue();
		Integer dataLength = convParameters.getDataLength();
		Integer numericPrecision = convParameters.getNumericPrecision();
		Integer numericScale = convParameters.getNumericScale();
		if (columnType.startsWith("VARCHAR")) {
			return newColumnType(Typ.VARCHAR, dataLength, defaultValue);
		} else if ("CHAR".equals(columnType)) {
			return newColumnType(Typ.CHAR, dataLength, defaultValue);
		} else if ("BLOB".equals(columnType)) {
			return newColumnType(Typ.BLOB, defaultValue);
		} else if ("CLOB".equals(columnType)) {
			return newColumnType(Typ.CLOB, defaultValue);
		} else if ("DATE".equals(columnType)) {
			return newColumnType(Typ.DATE, defaultValue);
		} else if ("TIMESTAMP".equals(columnType)) {
			return newColumnType(Typ.TIMESTAMP, defaultValue);
		} else if ("NUMBER".equals(columnType)) {
			return newColumnType(Typ.NUMBER, numericPrecision, numericScale, defaultValue);
		} else {
			return newColumnType(Typ.UNKNOWN);
		}
	}

	public static ColumnType convString(ConvParameters convParameters) {
		DatabaseType databaseType = convParameters.getDatabaseType();
		switch (databaseType) {
		case POSTGRES:
			return convPostgresString(convParameters);
		case ORACLE:
			return convOracleString(convParameters);
		}
		throw new DatabaseTypeException(databaseType);
	}

	private static String prepareType(String name, Integer precision) {
		return prepareType(name, (precision != null && precision > 0) ? precision.toString() : null);
	}

	private static String prepareType(String name, String precision) {
		StringBuilder sb = new StringBuilder();
		sb.append(name);
		if (precision != null && !precision.isEmpty()) {
			sb.append("(");
			sb.append(precision);
			sb.append(")");
		}
		return sb.toString();
	}

	public static String preparePrecision(Integer numericPrecision, Integer numericScale) {
		StringBuilder sb = new StringBuilder();
		if (numericPrecision != null && numericPrecision > 0) {
			sb.append(numericPrecision);
			if (numericScale != null) {
				sb.append(COMMA);
				sb.append(numericScale);
			}
		}
		return sb.toString();
	}

	private static String convPostgresTyp(ColumnType columnType) {
		switch (columnType.getType()) {
		case UNKNOWN:
			return "";
		case CLOB:
			return "text";
		case VARCHAR:
			return prepareType("varchar", columnType.getDataLength());
		case BLOB:
			return "bytea";
		case DATE:
			return "date";
		case TIMESTAMP:
			return "timestamp";
		case NUMBER:
			return prepareType("numeric",
					preparePrecision(columnType.getNumericPrecision(), columnType.getNumericScale()));
		case CHAR:
			return prepareType("char", columnType.getDataLength());
		}
		throw new ColumnTypeException(columnType);
	}

	private static Integer prepareVarcharDataLength(ColumnType type) {
		Integer dataLength = type.getDataLength();
		boolean isPrecision = (dataLength != null && dataLength > 0);
		if (isPrecision) {
			return dataLength;
		} else {
			return MAX_LENGHT_VARCHAR;
		}
	}

	private static String convOracleVarcharTyp(ColumnType type) {
		return prepareType("varchar2", prepareVarcharDataLength(type));
	}

	private static String convOracleTyp(ColumnType columnType) {
		switch (columnType.getType()) {
		case UNKNOWN:
			return "";
		case NUMBER:
			return prepareType("number",
					preparePrecision(columnType.getNumericPrecision(), columnType.getNumericScale()));
		case CLOB:
			return "clob";
		case VARCHAR:
			return convOracleVarcharTyp(columnType);
		case BLOB:
			return "blob";
		case DATE:
			return "date";
		case TIMESTAMP:
			return "timestamp";
		case CHAR:
			return prepareType("char", columnType.getDataLength());
		}
		throw new ColumnTypeException(columnType);
	}

	public static String convTyp(ColumnType type, DatabaseType databaseType) {
		switch (databaseType) {
		case POSTGRES:
			return convPostgresTyp(type);
		case ORACLE:
			return convOracleTyp(type);
		}
		throw new DatabaseTypeException(databaseType);
	}

}
